from cliriculum.renderers import Renderer
from cliriculum.markdown import ParseMd, LogoEntry
from cliriculum.deserializers import Dates, Contact
from cliriculum.loaders import load_json


def test_parse_md(fixtures_path):
    parsed = ParseMd(str(fixtures_path / "main.md"))
    dates_d = load_json(str(fixtures_path / "dates.json"))
    dates = Dates(dates_d)
    contact_d = load_json(str(fixtures_path / "contact.json"))
    contact = Contact(**contact_d)
    parsed.add_dates(dates=dates)
    parsed.add_contact(contact=contact)


def test_renderer(fixtures_path):
    """_summary_
    Test calling only.
    No result control
    Parameters
    ----------
    fixtures_path : _type_
        _description_
    """
    md_path = str(fixtures_path / "main.md")
    parsed = ParseMd(md_path)
    dates_d = load_json(str(fixtures_path / "dates.json"))
    dates = Dates(dates_d)
    doc = parsed.add_dates(dates=dates).doc
    with Renderer() as r:
        html = r.render(doc)
    contact_d = load_json(str(fixtures_path / "contact.json"))
    contact = Contact(**contact_d)
    doc = parsed.add_contact(contact=contact).doc
    with Renderer() as r:
        html = r.render(doc)



from typing import Union, List, Iterable
from pathlib import PosixPath, Path
from shutil import copy2
from importlib import resources
import os


def _check_dir(directory, overwrite):
    if os.path.isdir(directory):
        os.makedirs(directory, exist_ok=overwrite)
    elif os.path.isfile(directory) is False:
        os.mkdir(directory)
    else:
        raise ValueError("The specified file is not a directory and exists")


def get_resources_nodes(
    root, tree, resrcs: List, skip_dirs: Iterable = {"__pycache__"}
) -> List[PosixPath]:
    """
    Get resources nodes recursively.

    Parameters
    ----------
    root: root Traversable
        Directory containing the module
        Generally obtained with:
        `root=resources.files(package_name).parent`
        package_name, being the package installed on your system
    tree : Traversable
        _description_
    resrcs:
        Resources should be empty list for first call.
        Expands on each recursive call.
    skip_dirs: Iterable
        Directories in which resources should not be fetched.
        Default = {'__pycache__'}

    Returns
    -------
    List[PosixPath]
        List of nodes which are resources

    Examples
    --------
    >>> a = resources.files("cliriculum.data")
    >>> get_resources_nodes(a.parent.parent, a)

    """
    package = ".".join(tree.relative_to(root).parts)  # init
    for node in tree.iterdir():
        if node.is_dir() and node.parts[-1] not in set(skip_dirs):
            get_resources_nodes(root=root, tree=node, resrcs=resrcs)
        else:
            if (
                resources.files(package).joinpath(node).is_file()
                and node.parts[-1] != "__init__.py"
            ):
                resrcs.append(node)
    return resrcs


def copy_resources(directory: str, resource_root: str = "cliriculum.data"):
    """
    Copy resources to specified directory

    Parameters
    ----------
    directory: str Where to store the directory.
        The directory should exist.
    resource_root: str
        The resource_root directory specified in module type
    notation, default "cliriculum.data"
    """
    if os.path.isdir(directory) is False and os.path.isfile(directory) is False:
        raise FileNotFoundError(
            f"Directory: '{directory}' does not exist, it should be created first."
        )
    elif os.path.isdir(directory) is False and os.path.isfile(directory) is True:
        raise FileExistsError("File exists: '{directory}'")
    pkgname = resource_root.split(".")[0]

    tree = resources.files(package=resource_root)  # requires python >=3.9
    root = resources.files(package=pkgname).parent
    rsrcs = get_resources_nodes(root=root, tree=tree, resrcs=[])

    for source in rsrcs:
        relative_to = source.relative_to(tree)
        target_dir = Path(directory)
        target = target_dir / str(relative_to)

        # if source.relative_to is not root
        # and directory does not exist
        # create directory
        if target.parent.exists() is False and target.parent != target_dir:
            os.makedirs(target.parent)
        copy2(src=source, dst=target)
        # assert False == True


def copy_files(srcs: Iterable[Union[str, Path]], dst: Union[Path, str]) -> None:
    """
    Copy files to destination keeping path basename.

    Parameters
    ----------
    srcs : Iterable[Union[str, Path]]
        _description_
    dst : Union[Path, str]
        _description_
    """
    # effectivement peut etre très dangereux.
    # Doit etre corriger.
    if isinstance(srcs, str):
        raise TypeError("srcs should be an Iterable but not a string")
    for src in srcs:
        file_dst = Path(dst) / os.path.basename(src)
        copy2(src, file_dst)


def _case(string, camel_case: bool) -> str:
    if camel_case is True:
        string = string.capitalize()
    else:
        string = string.lower()
    return string


def _replace_space(string, camel_case):
    return "_".join(_case(token, camel_case=camel_case) for token in string.split(" "))


def auto_filename(name: str, job_title: str, company: str, camel_case: bool) -> str:
    """
    Create filename string under the form:

    `<name>-resume-<job title>-<company>.pdf` if camel_case is False.
    Or `<Name>-Resume-<Job title>-<Company>.pdf` if camel_case is True.

    Note that spaces contained in name will be replace by "_"

    Example
    -------
    No space in name:

    >>> from cliriculum.utils import auto_filename
    >>> auto_filename(name="Cliriculum", job_title="Resume Maker", "company"="No Company")
        'cliriculum-resume_maker-company'


    Space in name:

    >>> auto_filename(name="Cliriculum Tool", job_title="Resume Maker", "company"="No Company")
        'cliriculum_tool-resume_maker-company'

    Camel case in name:

    >>> auto_filename(name="Cliriculum Tool", job_title="Resume Maker", "company"="No Company", camel_case=True)
        'Cliriculum_Tool-Resume-Maker-Company'

    """

    return (
        _replace_space(name, camel_case=camel_case)
        + "-"
        + _case("resume", camel_case=camel_case)
        + "-"
        + _replace_space(job_title, camel_case=camel_case)
        + "-"
        + _replace_space(company, camel_case=camel_case)
        + ".pdf"
    )



"""
Utilities to render the generated resume from html to pdf.
"""

from shutil import which
import subprocess
from http.server import HTTPServer, SimpleHTTPRequestHandler
from typing import Union
from threading import Thread
import os


ARGS = ["--headless"]
EXTRA_ARGS = [
    "--run-all-compositor-stages-before-draw",
    "--disable-gpu",
    "--no-first-run",
    "--no-default-browser-check",
    "--hide-scrollbars",
]


def check_dependency(entrypoints):
    """
    Check if a dependency command line is available.


    Parameters
    ----------
    entrypoints: List[str]
        A list of synonym command line entry points.

    """
    value = False
    for entrypoint in entrypoints:
        cmd = which(entrypoint)
        if cmd is not None:
            value = entrypoint
            break
    return value


class DependencyError(Exception):
    pass


# Rather than relying on metaclass, which would allow setting the directory parameter dynamically by creating multiple classes as in
# https://stackoverflow.com/questions/39801718/how-to-run-a-http-server-which-serves-a-specific-path
# I chose just setting a context manager which changes the current directory.
# This choice might change to metaclass, if more evidence of benefits of this method appears in the future
# Current choice, avoids setting directory which is an argument added in Python 3.7 only


class ChDir:
    """
    Context manager to temporarily change directory.

    Example
    -------
    Use with http server
    Imagine an index.html is in the resume folder.
    >>> from http.server import HTTPServer, SimpleHTTPRequestHandler
    >>> from cliriculum.pdf import ChDir
    >>> with ChDir("resume/"):
    >>>     httpd = HTTPServer(('', 8000), SimpleHTTPRequestHandler)
    >>>     httpd.serve_forever()
    """

    def __init__(self, target: Union[str, os.PathLike]):
        # current dir
        # Checks
        if isinstance(target, (os.PathLike, str)) is False:  # check type
            raise TypeError("Only os.PathLike or str types are accepted.")
        # Convert to str absolute path

        target = os.path.abspath(target)  # pathlike automatically converted to str
        self.origin = os.getcwd()
        self.target = target

    def __enter__(self):
        os.chdir(self.target)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        os.chdir(self.origin)


def chromium_print(
    directory: Union[str, os.PathLike],
    filename: str = "output.pdf",
    port=8000,
    virtual_time_budget=20000,
):
    """

    Parameters
    ----------
    directory : Union[str, os.PathLike]
        Directory of the HTML resume
    filename : str, optional
        Name of the converted document, by default "output.pdf"
        Passed to headless chromium.
        **Note**: Can also be a a path (abs or relative).
    port : int, optional
        The network port , by default 8000
        If you wish automatic port selection set this value to 0.
    virtual_time_budget : int, optional
        Used to set chromium headless --virtual-time-budget flag, by default 20000

    Raises
    ------
    DependencyError
    subprocess.SubprocessError
    """
    cli_deps = ["chromium", "chromium-browser", "google-chrome", "google-chrome-stable"]
    ch_deps = check_dependency(cli_deps)
    if ch_deps is False:
        raise DependencyError
    with ChDir(directory):
        httpd = HTTPServer(("", port), SimpleHTTPRequestHandler)
        thread_server = Thread(target=httpd.serve_forever)
        thread_server.start()
        PRINT = [f"--print-to-pdf={filename}"]
        BUDGET = [f"--virtual-time-budget={virtual_time_budget}"]
        process = subprocess.run(
            [ch_deps]
            + ARGS
            + PRINT
            + BUDGET
            + EXTRA_ARGS
            + [f"http://127.0.0.1:{port}"]
        )
        if process.returncode != 0:
            raise subprocess.SubprocessError(f"return code: {process.returncode}")
        httpd.shutdown()



from mistletoe.block_token import (
    BlockCode,
    Heading,
    Quote,
    CodeFence,
    ThematicBreak,
    List,
    Table,
    Footnote,
    HTMLBlock,
    Paragraph,
    span_token,
)
from mistletoe.block_tokenizer import tokenize


# override mistletoe.block_token.Document defaults.
__all__ = [
    "BlockCode",
    "Heading",
    "Quote",
    "CodeFence",
    "ThematicBreak",
    "List",
    "Table",
    "Footnote",
    "HTMLBlock",
    "Paragraph",
]
_token_types = [globals()[cls_name] for cls_name in __all__]


class Document:
    """
    Copied from [mistletoe.block_token.Document](https://github.com/miyuchina/mistletoe/blob/276a7ec0f3ac541d93cec1fdb2fef582e48d9956/mistletoe/block_token.py#L136-L151)
    Instance attribute children is modified. I decided to copy instead of inherit the class to fix a scope issue.

    A copy of the original mistletoe license is available down below:

    The MIT License

    Copyright 2017 Mi Yu

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in
    the Software without restriction, including without limitation the rights to
    use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """

    token_types = _token_types

    def __init__(self, lines):
        if isinstance(lines, str):
            lines = lines.splitlines(keepends=True)
        lines = [line if line.endswith("\n") else "{}\n".format(line) for line in lines]
        self.footnotes = {}
        global _root_node
        _root_node = self
        span_token._root_node = self
        self.children = tokenize(lines, self.token_types)
        span_token._root_node = None
        _root_node = None



import json


def load_json(path):
    with open(path, mode="r") as f:
        text = json.load(f)
    return text


def load_css(path):
    with open(path, mode="r") as f:
        text = f.read()
    return text



from mistletoe.html_renderer import HTMLRenderer
from cliriculum.markdown import (
    LogoEntry,
    TextEntry,
    URLEntry,
    ContactBlock,
    SocialBlock,
    DescriptionBlock,
    ImageEntry,
    LocationEntry,
    PeriodEntry,
)
from textwrap import dedent


class Renderer(HTMLRenderer):
    def __init__(self):
        super().__init__(
            DescriptionBlock,
            ContactBlock,
            SocialBlock,
            TextEntry,
            URLEntry,
            ImageEntry,
            LocationEntry,
            PeriodEntry,
        )

    @staticmethod
    def _image_html(token):
        if token.width is None:
            width = ""
        else:
            width = 'width="{}"'.format(token.width)
        if token.height is None:
            height = ""
        else:
            height = 'height="{}"'.format(token.height)
        identifier = ""
        if hasattr(token, "id"):
            if token.id is not None:
                identifier = 'id="{}"'.format(token.id)

        html = '<img src="{src}" {width} {height} {identifier}/>'
        return html.format(
            src=token.src, width=width, height=height, identifier=identifier
        )

    @staticmethod
    def _i_image_html(token):
        template = '<i class="{classes}"></i>'
        return template.format(classes=token.classes)

    def render_logo_entry(self, token):
        text = token.title
        parent_o = '<div class="box">'
        parent_c = "</div>"

        if token.classes is not None:
            html = "\n".join(
                [parent_o, self._i_image_html(token), f"<span>{text}</span>", parent_c]
            )
        else:
            if token.src is None:
                html = "\n".join([parent_o, f"<span>{text}</span>", parent_c])
            else:
                html = "\n".join(
                    [
                        parent_o,
                        self._image_html(token),
                        f"<span>{text}</span>",
                        parent_c,
                    ]
                )
        return dedent(html)

    def render_period_entry(self, token):
        return self.render_logo_entry(token)

    def render_location_entry(self, token):
        return self.render_logo_entry(token)

    def render_image_entry(self, token):
        if token.src is not None:
            return self._image_html(token)
        else:
            return ""

    def render_text_entry(self, token):
        pre = "<p>"
        post = "</p>"
        if token.emphasis == "bold":
            html = """
            <strong>{text}</strong>
            """.format(
                text=token.text
            )
        elif token.emphasis == "italic":
            html = """
            <i>{text}</i>
            """.format(
                text=token.text
            )
        else:
            html = """
            <p>{text}</p>
            """.format(
                text=token.text
            )
        return pre + dedent(html) + post

    def render_url_entry(self, token):
        parent_o = '<div class="box">'
        parent_c = "</div>"
        url = token.url
        if token.text is None and token.url is not None:
            text = token.url
        elif token.text is None and token.url is None:
            return ""
        text = token.text
        if token.classes is not None:
            html = "\n".join(
                [
                    parent_o,
                    self._i_image_html(token),
                    f'<a href="{url}">{text}</a>',
                    parent_c,
                ]
            )
        else:
            if token.src is None:
                html = "\n".join([parent_o, f'<a href="{url}">{text}</a>', parent_c])

            else:
                html = "\n".join(
                    [
                        parent_o,
                        self._image_html(token),
                        f'<a href="{url}">{text}</a>',
                        parent_c,
                    ]
                )
        return dedent(html)

    def render_contact_block(self, token):
        template = """
        <div class="contact" id="contact">
        <div class="content">
        {inner}
        </div>
        </div>
        """

        return dedent(template.format(inner=self.render_inner(token)))

    def render_social_block(self, token):
        template = '<div class="social">\n{inner}\n</div>'
        inner = self.render_inner(token)
        return template.format(inner=inner)

    def render_description_block(self, token):
        template = (
            '<div class="description">\n<div class="content">\n{inner}\n</div>\n</div>'
        )
        inner = self.render_inner(token)
        return template.format(inner=inner)


class MainRenderer(Renderer):
    def render_document(self, token):
        """
        Renders document to html representation in `<main>...</main>` tags
        Parameters
        ----------
        token : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        template = """
        <main id = "main" class="main">
        <div class="content">
        {main}
        </div>
        </main>
        """
        return dedent(template.format(main=self.render_inner(token)))


class SideBarRenderer(Renderer):
    def render_document(self, token):
        """
        Renders document to html representation in `<aside>...</aside>` tags

        Parameters
        ----------
        token : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        template = '<aside id="aside" class="aside">\n{sidebar}\n</aside>'

        return template.format(sidebar=self.render_inner(token))



from cliriculum.utils import copy_files, copy_resources
import pytest
import os


def test_copy_files(tmp_path, fixtures_path):
    file_ = fixtures_path / "contact.json"
    with pytest.raises(TypeError):
        copy_files(file_, tmp_path)

    copy_files([file_], dst=tmp_path)


def test_copy_resources(tmp_path):
    target_dir = str(tmp_path / "resume")
    with pytest.raises(FileNotFoundError):
        copy_resources(target_dir)
    with open(target_dir, "w") as f:
        f.write("\n")
    with pytest.raises(FileExistsError):
        copy_resources(target_dir)
    os.remove(target_dir)
    os.mkdir(target_dir)
    copy_resources(target_dir)



from datetime import date
from typing import Union, List, Dict
from collections import UserDict


class Dates(UserDict):
    """
    Deserializer of date metadata

    Examples
    --------
    >>> d = load_json("dates.json")
    >>> dates = Dates(d)

    Attributes
    ----------
    periods: List[Periods]
    """

    def __setitem__(self, key, item):
        new_value = Period(id=key, **item)
        self.data[key] = new_value


class Period:
    """

    Attributes
    ----------
    start: str
    end: Union[str, None]
    idx: str
    logo: Union[str, None]
    width: Union[str, None]
    height: Union[str, None]
    classes: Union[str, None] classes has priority over logo in :py:mod:`cliriculum.renderers`
    """

    def __init__(
        self,
        id: str,
        start: date,
        end: Union[date, None] = None,
        logo: Union[str, None] = None,
        width: Union[str, None] = None,
        height: Union[str, None] = None,
        classes: Union[str, None] = None,
    ):
        """

        Parameters
        ----------
        id : str
        start : str
            The start of the period
        end : Union[str, None]
            The end of the period
        logo: Union[str, None]
            A path towards a logo
        width: Union[str, None]
            Width of logo
        height:  Union[str, None]
            Height of logo
        classes: Union[str, None]
            css classnames
            `classes="class1 class2"`
        """
        if start is None:
            raise ValueError(
                "`start` of {classname} can not be set to None".format(
                    classname=self.__class__.__name__
                )
            )
        start = date.fromisoformat(start)
        if end is not None:
            end = date.fromisoformat(end)

        self.start = start
        self.end = end
        self.idx = id
        self.logo = logo
        self.width = width
        self.height = height
        self.classes = classes

    def __str__(self):
        return f"Index: {self.idx} with start:{self.start} and end: {self.end}"


class URL:
    """
    Attributes
    ----------
    logo: str
    url: Union[str, None]
    classes: str
        See :py:class:`Period`
    text: str
    """

    def __init__(
        self,
        url: Union[str, None],
        logo: Union[str, None] = None,
        classes: Union[str, None] = None,
        text: Union[str, None] = None,
        width: Union[str, None] = None,
        height: Union[str, None] = None,
    ):
        self.logo = logo
        self.url = url
        self.classes = classes
        self.text = text
        self.width = width
        self.height = height


class Website(URL):
    pass


class Social(URL):
    pass


class Number(URL):
    pass


class Email(URL):
    pass


class Socials:
    """
    Attributes
    ----------
    children: List[Social]
    """

    def __init__(self, socials_list: List[Dict]):
        """
        Parameters
        ----------
        socials_dict : List[Dict]
            A list of dictionaries
            Passed to :py:class:`Social` as named arguments
        """

        self.children = [Social(**dict_) for dict_ in socials_list]


class Profile(URL):
    """
    Profile fields.

    Attributes:
    -----------
    ...: Attributes from: :py:class:`URL`
    """

    def __init__(
        self,
        picture: Union[str, None],
        width: Union[str, None] = "200px",
        height: Union[str, None] = "200px",
    ):
        """
        Parameters
        ----------
        picture : Union[str, None], optional
            Picture path, by default None
            passed as super(URL).__init__(logo=picture)
        width : Union[str, None], optional
            Width height, by default "200px"
        height : Union[str, None], optional
            Picture height, by default "200px"
        """
        super().__init__(url=None, width=width, height=height, logo=picture)


class Contact:
    """
    Contact deserializer.

    Attributes
    ----------
    name: str
    profession: str
    email: Email
    website: Website
    socials: Socials
    number: Number
    """

    def __init__(
        self,
        name,
        profession: Union[str, None] = None,
        email: Union[str, None] = None,
        website: Union[str, None] = None,
        socials: Union[str, None] = None,
        number: Union[str, None] = None,
        profile: Union[str, None] = None,
    ):
        """
        name:
            Required
            Single string (first name last name)
        profession: Union[str, None]
            Defaults to None.
        email: Union[str, None]
            Defaults to None.
        website: Union[str, None]
            Defaults to None.
        socials: Union[str, None]
            Defaults to None.
        number: Union[str, None]
            Defaults to None

        Examples
        --------
        >>> from cliriculum.parsers import load_json
        >>> c = load_json("contact.json")
        >>> Contact(**c)
        """
        self.name = name
        self.profession = profession
        # all classes subclassing URL require url parameter at least
        # i.e:
        # - Email
        # - Website
        # - Number
        if email is None:
            email = {"url": None}  # bypassing restriction
        if website is None:
            website = {"url": None}
        if socials is None:
            socials = []
        if profile is None:
            profile = {"picture": None}
        if number is None:
            number = {"url": None}

        self.email = Email(**email)
        self.website = Website(**website)
        self.socials = Socials(socials)

        self.number = Number(**number)
        self.profile = Profile(**profile)


class Location:
    def __init__(self, id: str, location: str, classes: Union[str, None] = None):
        """

        Parameters
        ----------
        id : str
            An id to match with content
        location : str
            _description_
        classes : Union[str, None], optional
            _description_, by default None
        """
        self.classes = classes
        self.location = location
        self.idx = id


class Locations(UserDict):
    """

    Parameters
    ----------
    UserDict : _type_

    Example
    -------
    >>> from cliriculum.parsers import load_json
    >>> from cliriculum.deserializers import Locations
    >>> l = load_json("location.json")
    >>> Locations(l)
    """

    def __setitem__(self, key, item):
        new_value = Location(id=key, **item)
        self.data[key] = new_value


class Job:
    def __init__(
        self, title: Union[str, None] = None, company: Union[str, None] = None
    ):
        self.title = title
        self.company = company



from cliriculum.markdown import ParseMd
from cliriculum.renderers import SideBarRenderer, MainRenderer
from cliriculum.deserializers import Contact, Dates, Locations
from cliriculum.markdown import DescriptionBlock
from cliriculum.loaders import load_json
from warnings import warn
from typing import Union
from cliriculum.utils import copy_files
from os.path import basename


HEAD = """
<!doctype html>
<head>
  <meta charset="UTF-8">
  <title>HTML Resume</title>
  <link rel="stylesheet" href="style.css" type="text/css">
  <link rel="stylesheet" href="fontawesome/css/all.css" type="text/css">
  <link rel="stylesheet" href="fontawesome/css/v4-shims.css" type="text/css">
  {additionalcss}
  <script src="paged.polyfill.js"></script>
  <script src="style.js"></script>
</head>
"""


class SideBarHTML:
    def __init__(self, path, contact, rsrc_dst: Union[str, None] = None):
        """_summary_

        Parameters
        ----------
        path : _type_
            _description_
        contact : _type_
            _description_
        rsrc_dst : Union[str, None], optional
            If not None resources mentionned in JSON files are copied to destination directory `rsrc_dst`. By default, None.
        """
        parsed = ParseMd(path)
        contact_d = load_json(contact)
        contact_o = Contact(**contact_d)
        parsed.add_contact(contact_o)
        doc = parsed.doc
        # find ...
        if parsed.top is True:
            contact = doc.children[0]
            description = doc.children[1:]
        else:
            contact = doc.children[-1]
            description = doc.children[:-1]
        # all except this element
        # either at beginning either at the end
        children = [contact, DescriptionBlock(children=description)]
        doc.children = children
        with SideBarRenderer() as r:
            self.html = r.render(doc)
        if rsrc_dst is not None and contact_o.profile.logo is not None:
            warn("Currently, specifying rsrc_dst copies only contact profile picture")
            copy_files([contact_o.profile.logo], rsrc_dst)
        self._contact = contact_o


class MainHTML:
    def __init__(
        self, path, dates: Union[str, None] = None, location: Union[str, None] = None
    ):
        self.parsed = ParseMd(path)
        if dates is not None:
            dates_d = load_json(dates)
            dates_o = Dates(dates_d)
            self.parsed.add_dates(dates_o)

        if location is not None:
            location_d = load_json(location)
            location_o = Locations(location_d)
            self.parsed.add_location(location_o)

        with MainRenderer() as r:
            self.html = r.render(self.parsed.doc)


class ResumeHTML:
    """
    Render Resume

    Attributes
    ----------
    main: MainHTML
    sidebar: SideBarHTML
    """

    def __init__(
        self,
        sidebar: SideBarHTML,
        main: MainHTML,
        stylesheet: Union[str, None],
        rsrc_dst=Union[str, None],
    ):
        """

        Parameters
        ----------
        sidebar : SideBarHTML
        main : MainHTML
        stylesheet: Path to stylesheet.
        rsrc_dst: See :class:SidebarHTML

        Returns
        -------
        _type_
            _description_
        """
        self.main = main
        self.sidebar = sidebar
        if stylesheet is not None and rsrc_dst is None:
            raise ValueError("stylesheet can not be used when rsrc_dst is set to None")
        if stylesheet is not None:
            copy_files(srcs=[stylesheet], dst=rsrc_dst)
            _ = f'<link rel="stylesheet" href="{basename(stylesheet)}" type="text/css">'
        else:
            _ = ""
        self.additionalcss = _

    def join(self) -> str:
        body = self.sidebar.html + "\n" + self.main.html  # union of sidebar and main
        template = "<html>{head}<body>{body}</body></html>"
        return template.format(
            head=HEAD.format(additionalcss=self.additionalcss), body=body
        )


def resume(
    sidebar_md: str,
    main_md: str,
    dates: str,
    contact: str,
    rsrc_dst: Union[str, None] = None,
    stylesheet: Union[str, None] = None,
    location: Union[str, None] = None,
) -> str:
    """

    Parameters
    ----------
    sidebar_md : str
        Path to sidebar markdown
    main_md : str
        Path to main markdown
    dates : str
        path to dates JSON
    contact : str
        Path to contact JSON
    rsrc_dst: Union[str, None]
        Argument passed to :class:SidebarHTML and :class:ResumeHTML
    stylesheet: Argument passed to :class:ResumeHTML
    location: Union[str, None]
        Path to JSON file containing location metadata

    Returns
    -------
    str
        HTML representation of the resume

    Examples
    --------
    >>> from cliriculum import resume
    >>> html = resume(sidebar_md="sidebar.md", main_md="main.md", contact="contact.json", dates="dates.json")
    """
    warn(
        "Deprecated in 0.1.6, Resume class should be used instead.", DeprecationWarning
    )
    sidebar = SideBarHTML(path=sidebar_md, contact=contact, rsrc_dst=rsrc_dst)
    main = MainHTML(path=main_md, dates=dates, location=location)
    resume = ResumeHTML(
        main=main, sidebar=sidebar, stylesheet=stylesheet, rsrc_dst=rsrc_dst
    )
    html = resume.join()
    return html


class Resume:
    """

    Attributes
    ----------
    stylesheet: str
    rsrc_dst: str
    resume: Union[ResumeHTML, None]
        Is set on __call__ (to ResumeHTML instance)

    Example
    -------
    >>> from cliriculum.resume import Resume
    >>> resume = Resume()  # a callable
    >>> resume(sidebar_md="sidebar.md", main_md="main.md", contact="contact.json", dates="dates.json")
    """

    def __init__(
        self,
        rsrc_dst: Union[str, None] = None,
        stylesheet: Union[str, None] = None,
    ):
        self.rsrc_dst = rsrc_dst
        self.stylesheet = stylesheet
        self.resume = None

    def __call__(
        self,
        sidebar_md: str,
        main_md: str,
        contact: str,
        dates: Union[str, None] = None,
        location: Union[str, None] = None,
    ) -> str:
        sidebar = SideBarHTML(path=sidebar_md, contact=contact, rsrc_dst=self.rsrc_dst)
        main = MainHTML(path=main_md, dates=dates, location=location)
        resume = ResumeHTML(
            main=main,
            sidebar=sidebar,
            stylesheet=self.stylesheet,
            rsrc_dst=self.rsrc_dst,
        )
        self.resume = resume  # wich means a text is generated
        html = resume.join()
        return html



from cliriculum.pdf import chromium_print, ChDir
from http.server import SimpleHTTPRequestHandler, HTTPServer
import os
import requests
from threading import Thread
from pypdf import PdfReader


def test_chdir(fixtures_path):
    with ChDir("example") as d:
        current = os.getcwd()
        assert current.find("example") >= 0
        assert d.target.find("example") >= 0

        assert d.target == current, f"target: {d.target}, current: {current}"
        httpd = HTTPServer(("", 8000), SimpleHTTPRequestHandler)
        t = Thread(target=httpd.serve_forever)
        t.start()
        # get html from httpd server
        with requests.get("http://127.0.0.1:8000") as r:
            assert (
                r.text.find("cliriculum") >= 0
            ), r.text  # check word cliriculum in request (i.e example/index.html
            # is effectively served)
        httpd.shutdown()
    current = os.getcwd()
    assert d.origin == current, f"origin: {d.origin}, current: {current}"


def test_chromium_print():
    chromium_print("example/", virtual_time_budget=20000)
    # check if pdf exists
    filep = "example/output.pdf"
    assert os.path.exists(filep), "file: {filep} does not exist"
    # parse pdf
    max_repeat = 2
    i = 0
    while i < max_repeat:
        chromium_print("example/", virtual_time_budget=20000)
        reader = PdfReader(filep)
        page = reader.pages[0]
        text = page.extract_text()
        text_length = len(text)
        if text_length > 0:
            break
        if text_length == 0 and i == max_repeat - 1:
            assert text_length > 0, f"test failed `max_repeat`: {max_repeat} times"
        i += 1



from cliriculum.utils import copy_resources, copy_files
from cliriculum.main import make_resume
from cliriculum.deserializers import Contact
from cliriculum.loaders import load_json
from pathlib import Path
import os
import pytest

BASE_DIR = Path(__file__).parent.parent


def _isnot_pycache_or_init(path: Path):
    if "__pycache__" in path.parts or "__init__.py" in path.parts:
        return False
    else:
        return True


def test_copy_resources(tmp_path):
    copy_resources(directory=tmp_path)
    lev1 = tmp_path.glob("*")
    lev2 = tmp_path.glob("*/*")
    lev3 = tmp_path.glob("*/*/*")
    expected_path = BASE_DIR / "cliriculum/data"

    rsrcs1 = list(filter(_isnot_pycache_or_init, expected_path.glob("*")))
    rsrcs2 = list(filter(_isnot_pycache_or_init, expected_path.glob("*/*")))
    rsrcs3 = list(filter(_isnot_pycache_or_init, expected_path.glob("*/*/*")))

    t1 = set([path.name for path in lev1]).symmetric_difference(
        [path.name for path in rsrcs1]
    )
    t2 = set([path.name for path in lev2]).symmetric_difference(
        [path.name for path in rsrcs2]
    )
    t3 = set([path.name for path in lev3]).symmetric_difference(
        [path.name for path in rsrcs3]
    )

    assert t1 == set()
    assert t2 == set()
    assert t3 == set()


def test_copy_files(tmp_path, fixtures_path):
    c = str(fixtures_path / "contact.json")
    contact = Contact(**load_json(c))
    copy_files(srcs=[contact.profile.logo], dst=tmp_path)
    img = "image.png"
    t1 = set(os.listdir(tmp_path)).symmetric_difference([img])
    assert t1 == set()


def test_make_resume(tmp_path, fixtures_path):
    s = str(fixtures_path / "sidebar.md")
    m = str(fixtures_path / "main.md")
    d = str(fixtures_path / "dates.json")
    c = str(fixtures_path / "contact.json")
    j = str(fixtures_path / "job.json")
    with pytest.raises(FileExistsError):
        make_resume(
            tmp_path,
            sidebar_md=s,
            main_md=m,
            dates=d,
            contact=c,
            overwrite=False,
            stylesheet=None,
        )

    make_resume(
        tmp_path,
        sidebar_md=s,
        main_md=m,
        dates=d,
        contact=c,
        overwrite=True,
        stylesheet=None,
    )
    assert "index.html" in os.listdir(str(tmp_path))

    make_resume(
        tmp_path,
        sidebar_md=s,
        main_md=m,
        dates=d,
        contact=c,
        overwrite=True,
        stylesheet=None,
        job_metadata=j,
        pdf_auto=True,
    )



from cliriculum.deserializers import Contact, Profile, Socials, URL
from cliriculum.deserializers import Dates, Period
from datetime import date, timedelta
import pytest


WEBSITE_1 = {
    "url": "https://github.com/sondalex/cliriculum",
    "logo": "github.com/sondalex/cliriculum",
    "classes": None,
    "text": None,
    "width": None,
    "height": None,
}
# fa-brands fa-linkedin"
WEBSITE_2 = {
    "url": "https://github.com/sondalex/cliriculum",
    "classes": "fa-brands fa-linkedin",
    "text": "github.com/sondalex/cliriculum",
    "width": "300px",
    "height": "300px",
}

KWARGS = [
    (
        {
            "profession": None,
            "email": None,
            "website": None,
            "socials": None,
            "number": None,
            "profile": None,
        },
        {
            "profession": "Data Scientist",
            "email": {"url": "test@email.com", "text": "test@email.com"},
            "website": WEBSITE_1,
            "socials": [WEBSITE_1],
            "number": WEBSITE_1,
            "profile": {"picture": "tests/fixtures/image.png"},
        },
        {
            "profession": "Data Scientist",
            "email": {"url": "test@email.com", "text": "test@email.com"},
            "website": WEBSITE_2,
            "socials": [WEBSITE_2],
            "number": WEBSITE_2,
        },
        {"profession": "Data Scientist"},
    )
]

PERIOD_KWARGS = [
    (
        {
            "start": date.today().isoformat(),
            "end": (date.today() + timedelta(1)).isoformat(),
            "logo": "github.com/sondalex/cliriculum",
            "width": "200px",
            "height": "200px",
            "classes": "fa-solid fa-location-pin",
        },
        {
            "start": date.today().isoformat(),
            "end": None,
            "logo": None,
            "width": None,
            "height": None,
            "classes": None,
        },
    )
]


class TestContact:
    @pytest.mark.parametrize("tuple", KWARGS)
    def test___init__(self, tuple):
        for dict_ in tuple:
            Contact("cliriculum", **dict_)
        with pytest.raises(TypeError):
            Contact()
        with pytest.raises(TypeError):
            Contact(**{})


class TestDates:
    @pytest.mark.parametrize("tuple", PERIOD_KWARGS)
    def test___init__(self, tuple):
        with pytest.raises(TypeError):
            Period(id="an-id")
        for dict_ in tuple:
            Period(id="an-id", **dict_)


class TestDates:
    @pytest.mark.parametrize("tuple", PERIOD_KWARGS)
    def test___init__(self, tuple):
        Dates({})  # not raise Error
        d = {f"an-id_{i}": item for i, item in enumerate(tuple)}
        Dates(d)



import argparse
import os
import sys
from pathlib import Path
from typing import Union
from cliriculum.resume import Resume
from cliriculum.utils import copy_resources, auto_filename, _check_dir
from cliriculum.pdf import chromium_print
from cliriculum.deserializers import Job
from cliriculum.loaders import load_json


def make_resume(
    directory,
    sidebar_md: str,
    main_md: str,
    dates: str,
    contact: str,
    overwrite: bool,
    stylesheet: Union[str, None],
    location: Union[str, None] = None,
    output: Union[str, None] = None,
    job_metadata: Union[str, None] = None,
    pdf_auto: bool = False,
):
    """
    Generates resume


    Parameters
    ----------
    ...:
        See :py:mod:`cliriculum.resume`
    directory: Directory to store the resume and all its file component (css, html, pdf, ...)
    overwrite: bool
        Whether to overwrite the `directory`
    output: Union[str, None]
        If provided a pdf is created in `directory + pdf`.
    pdf_auto:
        Precedence over output. Generates the pdf name automatically based on `job_metadata` if the latter is not None.
        Else defaults to output.pdf. By default False
    job_metadata: Union[str, None]
        JSON file location to job metadata, by default False.
    """
    _check_dir(directory=directory, overwrite=overwrite)
    resume = Resume(rsrc_dst=directory, stylesheet=stylesheet)
    html = resume(
        sidebar_md=sidebar_md,
        main_md=main_md,
        contact=contact,
        dates=dates,
        location=location,
    )

    with open(Path(directory) / "index.html", "w") as f:
        f.write(html)

    # copy dependencies
    print(f"Copying files to {directory}")
    copy_resources(directory)
    if output is not None and pdf_auto is False:
        chromium_print(directory=directory, filename=output)
    elif pdf_auto is True and job_metadata is not None:
        job_meta_dict = load_json(job_metadata)
        jobmeta = Job(**job_meta_dict)

        contact = resume.resume.sidebar._contact
        filename = auto_filename(
            name=contact.name,
            company=jobmeta.company,
            job_title=jobmeta.title,
            camel_case=False,
        )
        chromium_print(directory=directory, filename=filename)
    elif pdf_auto is True and job_metadata is None:
        raise ValueError("If `pdf_auto` is set to True job metadata must be set. ")


def cli() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Create a Resume using JSON and Markdown"
    )
    parser.add_argument(
        "--main",
        dest="main_md",
        type=str,
        action="store",
        required=True,
        help="Path to the markdown file",
    )
    parser.add_argument(
        "--description", dest="descr_md", type=str, action="store", required=True
    )
    parser.add_argument(
        "--dates",
        dest="dates_json",
        type=str,
        action="store",
        required=True,
        help="Path to the dates JSON file",
    )
    parser.add_argument(
        "--contact",
        dest="contact_json",
        type=str,
        action="store",
        required=True,
        help="Path to the contact JSON file",
    )

    parser.add_argument(
        "--destination",
        dest="destination",
        type=str,
        action="store",
        required=False,
        help="Directory where the HTML resume should be created",
        default="resume",
    )

    parser.add_argument(
        "--overwrite",
        type=int,
        action="store",
        required=False,
        help="Overwrite existing directory if set to other value than 0.",
        default=0,
    )
    parser.add_argument(
        "--stylesheet",
        dest="stylesheet",
        type=str,
        action="store",
        required=False,
        help="A supplementary stylesheet to add to the dom.",
        default=None,
    )
    parser.add_argument(
        "--location",
        dest="location",
        type=str,
        action="store",
        required=False,
        help="A file specifying locations to map to 2nd level headers by id",
        default=None,
    )
    parser.add_argument(
        "--pdf-output",
        dest="output",
        action="store",
        type=str,
        default=None,
        required=False,
        help="If filename the pdf will be generated and saved to '--directory + --pdf-output', if path like it will be saved to provided path",
    )
    parser.add_argument(
        "--pdf-auto",
        dest="pdf_auto",
        type=int,
        default=0,
        required=False,
        help="If not 0, the pdf will be named automatically based on job-metadata. If job-metadata is not provided it will default to 'output.pdf'. Precedence over pdf-output",
    )
    parser.add_argument(
        "--job-metadata",
        dest="job_metadata",
        action="store",
        type=str,
        default=None,
        required=False,
        help="Path to json file including metadata about job",
    )
    return parser


def main():
    parser = cli()
    args = parser.parse_args()
    overwrite = bool(args.overwrite)
    make_resume(
        directory=args.destination,
        sidebar_md=args.descr_md,
        contact=args.contact_json,
        dates=args.dates_json,
        main_md=args.main_md,
        overwrite=overwrite,
        stylesheet=args.stylesheet,
        location=args.location,
        output=args.output,
        job_metadata=args.job_metadata,
        pdf_auto=bool(args.job_metadata),
    )


if __name__ == "__main__":
    sys.exit(main())



from cliriculum.resume import resume, SideBarHTML, MainHTML, ResumeHTML, Resume
import pytest
import os
import warnings


ARGS = ["location_path", "location_none"]


class TestResume:
    @pytest.mark.parametrize("location", ARGS)
    def test___init__and__call__(self, location, tmp_path, fixtures_path, request):
        s = str(fixtures_path / "sidebar.md")
        m = str(fixtures_path / "main.md")
        d = str(fixtures_path / "dates.json")
        c = str(fixtures_path / "contact.json")
        add_css = str(fixtures_path / "custom.css")
        resume = Resume(rsrc_dst=None, stylesheet=None)
        html = resume(
            sidebar_md=s,
            main_md=m,
            dates=d,
            contact=c,
            location=request.getfixturevalue(location),
        )
        assert len(html) > 0


def test_sidebar_html(fixtures_path):
    s = str(fixtures_path / "sidebar.md")
    c = str(fixtures_path / "contact.json")

    sidebar = SideBarHTML(path=s, contact=c)


@pytest.mark.parametrize("location", ARGS)
def test_main_html(location, fixtures_path, request):
    m = str(fixtures_path / "main.md")
    d = str(fixtures_path / "dates.json")
    main = MainHTML(m, dates=d, location=request.getfixturevalue(location))


# This test tests deprecated function (depr from 0.1.6)
# -----------------------------------
@pytest.mark.parametrize("location", ARGS)
def test_resume(location, tmp_path, fixtures_path, request):
    s = str(fixtures_path / "sidebar.md")
    m = str(fixtures_path / "main.md")
    d = str(fixtures_path / "dates.json")
    c = str(fixtures_path / "contact.json")
    add_css = str(fixtures_path / "custom.css")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        html = resume(sidebar_md=s, main_md=m, dates=d, contact=c)
    with pytest.deprecated_call():
        resume(sidebar_md=s, main_md=m, dates=d, contact=c)


@pytest.mark.parametrize("location", ARGS)
def test_resume_html(location, tmp_path, fixtures_path, request):
    s = str(fixtures_path / "sidebar.md")
    m = str(fixtures_path / "main.md")
    d = str(fixtures_path / "dates.json")
    c = str(fixtures_path / "contact.json")
    sidebar = SideBarHTML(path=s, contact=c, rsrc_dst=tmp_path)
    main = MainHTML(path=m, dates=d, location=request.getfixturevalue(location))
    resume = ResumeHTML(
        main=main,
        sidebar=sidebar,
        stylesheet=fixtures_path / "custom.css",
        rsrc_dst=tmp_path,
    )
    assert resume.additionalcss != ""

    files = os.listdir(tmp_path)
    assert set(files).symmetric_difference(["custom.css", "image.png"]) == set()

    with pytest.raises(ValueError):
        ResumeHTML(
            main=main,
            sidebar=sidebar,
            stylesheet=fixtures_path / "custom.css",
            rsrc_dst=None,
        )



import re
from mistletoe.block_token import Heading
from mistletoe.span_token import RawText
from cliriculum.deserializers import Dates, Contact, Locations
from typing import List, Union, Type, Any, Tuple
import os
from cliriculum.parsers import Document
from collections import OrderedDict


class URLEntry:
    def __init__(self, src, width, height, url, classes, text):
        self.src = src
        self.url = url
        self.width = width
        self.height = height
        self.classes = classes
        self.text = text


class LogoEntry:
    def __init__(self, src, title, classes, width="18", height="18"):
        """
        A node intented to be added into the abstract
        syntax tree.
        The node contains the following attributes:

        src:
        title:
        width:
        height:

        Parameters
        ----------
        src : _type_
            Path to logo
        title : _type_
            A description
        width : str, optional
            Size of width of image, by default "18"
        height : str, optional
            Size of height of image, by default "18"
        """
        self.src = src
        self.title = title
        self.width = width
        self.height = height
        self.classes = classes

    def __str__(self):
        return f"< Instance of type: {self.__class__.__name__} with attributes src={self.src}, title={self.title}, width={self.width}, height={self.height}, classes=({self.classes})>"


class PeriodEntry(LogoEntry):
    def __init__(self, idx, start, end, logo, width, height, classes):
        # Bridge from cliriculum.deserializers.Period instance attributes
        # to LogoEntry
        startfmt = start.strftime("%B %Y")
        if end is not None:
            endfmt = end.strftime("%B %Y")
        else:
            endfmt = "Current"
        title = f"{startfmt} - {endfmt}"
        super().__init__(
            src=None, title=title, classes=classes, height=None, width=None
        )


class LocationEntry(LogoEntry):
    def __init__(self, idx, classes, location):
        # Bridge from cliriculum.deserializers.Location instance attributes
        # to LogoEntry
        title = location
        super().__init__(
            src=None, title=title, classes=classes, height=None, width=None
        )


class ImageEntry:
    def __init__(self, src, width, height, id):
        self.id = id
        self.src = src
        self.width = width
        self.height = height


class TextEntry:
    def __init__(self, text, emphasis: Union[str, None] = None):
        """_summary_

        Parameters
        ----------
        text : _type_
            _description_
        emphasis : str
            s
        """
        self.text = text
        _ = ["bold", "italic", None]
        if emphasis not in _:
            raise ValueError("`emphasis` accepts values : {}".format(_))
        self.emphasis = emphasis


class SocialBlock:
    def __init__(self, urlentries: List[URLEntry]):
        self.children = urlentries


class ContactBlock:
    def __init__(
        self, children: List[Union[SocialBlock, URLEntry, LogoEntry, TextEntry]]
    ):
        self.children = children


class DescriptionBlock:
    def __init__(self, children: List):
        self.children = children


class ParseMd:
    """
    Parse Markdown Document and adds
    metadata to the tree depending on method called.

    Attributes
    ----------
    doc The tree representation of the document. Each public methods calls
    modifies this attribute
    top: Union[bool, None]
        Set on `self.add_contact` calls.
        If multiple calls, the top correspond
        to the last position of the last
        added Contact node.
        Default None.
    """

    heading_id_pattern = re.compile(r"\{#([\w-]*[\w-])\}$")

    def __init__(self, path: str):
        """_summary_

        Parameters
        ----------
        path : str
            Path to markdown file.
            The file will be parsed by [mistletoe](https://github.com/miyuchina/mistletoe)
            And its tree representation stored in `self.doc`.
        """
        with open(path, mode="r") as f:
            self.doc = Document(f)
        self.periods = []
        self.top = None
        self.heading_index = OrderedDict()  # maybe should set it to private

    def _isHeading(self, leaf):
        if isinstance(leaf, Heading):
            return True

    def _headings_with_id_idx(self, doc: Document) -> OrderedDict:
        """_summary_

        Parameters
        ----------
        doc : Document
            _description_

        Returns
        -------
        OrderedDict
            An ordered dict with key corresponding to position in first tree layer
            and value the string identifier extracted from the header.
        """
        heading_index = OrderedDict()
        for i, leaf in enumerate(doc.children):
            # issubclass
            # isinstance
            if isinstance(leaf, Heading):
                outer_and_inner_span = self._get_heading_id_spans(leaf)
                if (leaf.level == 2) and (outer_and_inner_span is not None):
                    raw_leaf = leaf.children[0]
                    _, indice_span = outer_and_inner_span
                    indice = raw_leaf.content[indice_span[0] : indice_span[1]]
                    heading_index[i] = indice
        return heading_index

    def _get_heading_id_spans(
        self, node
    ) -> Union[Tuple[Tuple[int, int], Tuple[int, int]], None]:
        """_summary_

        Parameters
        ----------
        node : _type_
            _description_

        Returns
        -------
        Union[Tuple[str, str], None]
            If match returns the outer and inner span of match.

        Example
        -------
        >>> self._get_heading_id_spans(node)
        >>> ((7, 15), (9, 14))

        Raises
        ------
        ValueError
            _description_
        """
        if self._isHeading(node):
            raw_leaf = node.children[0]
            match_obj = self.heading_id_pattern.search(raw_leaf.content)
            if match_obj is not None:
                return match_obj.regs
            else:
                return None
        else:
            raise ValueError("Wrong type of node")

    def _has_heading_id(self, node):
        if self._isHeading(node):
            raw_leaf = node.children[0]  # select raw content
            if isinstance(raw_leaf, RawText) is False:
                error = """
                There might be an issue with the current version,
                could you report it ?
                """
                raise ValueError(error)

            match_obj = self.heading_id_pattern.search(raw_leaf.content)
            if match_obj is None:
                return False
            else:
                return True

        else:
            raise ValueError("Wrong type of node")

    def _add_to_top_paragraph(self, leveltwo_header_idx, new_node):
        """_summary_
        Add LogoEntry node just below `leveltwo_header_idx` (Second level with id's node index)

        Parameters
        ----------
        leveltwo_header_idx : Integer
            The index of the level two Heading two which Dates should be added
        """
        self.doc.children.insert(leveltwo_header_idx + 1, new_node)

    def add_node_by_match_idx(
        self, nodes: Union[Locations, Dates], new_node_class: Type[Any]
    ):
        """
        General method for adding node by matching
        idx. The matching is made on secondary level headers.

        Parameters
        ----------
        nodes: Union[Locations, Dates]
            Object derived from collections.UserDict class.
        new_node_class: ...
            A Class callable.
            On node match, the attributes of the matched node(s) from `nodes`
            are passed to new_node_class instanciation.

        Important Note
        --------------
        If you wish to render the parsed document, you have to make sure all `new_node_class` are registered
            in the renderer.
        """
        if len(self.heading_index) == 0:
            self.heading_index = self._headings_with_id_idx(self.doc)
            replace = True
        else:
            replace = False

        for i in range(len(self.heading_index)):
            pos = list(self.heading_index.keys())[i]
            # Match with Dates
            if replace:  # replace on first call
                heading = self.doc.children[pos]
                # extract id
                outer_span, inner_span = self._get_heading_id_spans(heading)
                raw_leaf = heading.children[0]
                indice = raw_leaf.content[inner_span[0] : inner_span[1]]
                to_replace = raw_leaf.content[outer_span[0] : outer_span[1]]
                self.doc.children[pos].children[0].content = raw_leaf.content.replace(
                    to_replace, ""
                ).rstrip()  # replace the {#id} with empty string
            else:
                indice = self.heading_index[pos]
            try:
                node = nodes[indice]
            except KeyError:
                node = None

            if node is not None:
                node_attributes = (
                    name for name in dir(node) if not name.startswith("_")
                )
                dictionnary = {key: getattr(node, key) for key in node_attributes}
                new_node = new_node_class(**dictionnary)
                self._add_to_top_paragraph(
                    leveltwo_header_idx=pos, new_node=new_node
                )  # operates directly on AST

            # Correct registered indices for addition of node
            h_index_items = list(self.heading_index.items())
            self.heading_index = OrderedDict(
                [(key, value) for key, value in h_index_items[0 : i + 1]]
                + [(key + 1, value) for key, value in h_index_items[i + 1 :]]
            )

        return self

    def add_dates(self, dates: Dates):
        """_summary_
        Adds dates as a LogoEntry to the doc representation


        Parameters
        ----------
        dates : Dates
            Instanciated Dates object.

        Returns
        -------
        mistletoe.Document
            An extended document with date added for Dates.periods object
            who share a level two ATX heading common id.

        Examples
        -------

        >>> from cliriculum.renderers import Renderer
        >>> from cliriculum.deserializers import Dates
        >>> from cliriculum.markdown import ParseMd
        >>> from cliriculum.loaders import load_json
        >>> parsed = ParseMd("README.md")
        >>> d = load_json("dates.json")
        >>> dates = Dates(d)
        >>> doc = parsed.add_dates(dates=dates).doc
        >>> with Renderer() as r:
        >>>     html = r.render(doc)
        """
        self.add_node_by_match_idx(nodes=dates, new_node_class=PeriodEntry)
        return self

    def add_contact(self, contact: Contact, top: bool = True):
        """_summary_

        Parameters
        ----------
        contact : Contact
            _description_
        top : bool, optional
            _description_, by default True

        Examples
        --------

        >>> from cliriculum.renderers import Renderer
        >>> from cliriculum.deserializers import Contact
        >>> from cliriculum.loaders import load_json
        >>> parsed = ParseMd("README.md")
        >>> dates = Contact(**load_json("contact.json"))
        >>> doc = parsed.add_contact(contact=contact).doc
        >>> with Renderer() as r:
        >>>     html = r.render(doc)
        """
        self.top = top
        name = TextEntry(contact.name, emphasis="bold")
        profession = TextEntry(contact.profession, emphasis="italic")
        email_url = contact.email.url
        if email_url is not None:
            email_url = "mailto: {}".format(contact.email.url)

        email = URLEntry(
            src=contact.email.logo,
            width=contact.email.width,
            height=contact.email.height,
            url=email_url,
            classes=contact.email.classes,
            text=contact.email.text,
        )
        website = URLEntry(
            src=contact.website.logo,
            width=contact.website.width,
            height=contact.website.height,
            url=contact.website.url,
            classes=contact.website.classes,
            text=contact.website.text,
        )
        socials = SocialBlock(
            [
                URLEntry(
                    src=social.logo,
                    width=social.width,
                    height=social.height,
                    url=social.url,
                    classes=social.classes,
                    text=social.text,
                )
                for social in contact.socials.children
            ]
        )
        if contact.number.url is None:
            tel_url = contact.number.url
        else:
            tel_url = "tel:{}".format(contact.number.url)
        number = URLEntry(
            src=contact.number.logo,
            width=contact.number.width,
            height=contact.number.height,
            url=tel_url,
            classes=contact.number.classes,
            text=contact.number.text,
        )
        children = [name, profession, email, website, socials, number]

        if contact.profile.logo is not None:
            profile = ImageEntry(
                src=os.path.basename(contact.profile.logo),
                height=contact.profile.height,
                width=contact.profile.width,
                id="profile_pic",
            )
            children = [profile] + children

        contact_block = ContactBlock(children=children)

        if top is False:
            self.doc.children.append(contact_block)
        else:
            self.doc.children.insert(0, contact_block)

            # increment all keys by 1
            self.heading_index = {
                key + 1: value for key, value in self.heading_index.items()
            }

        return self

    def add_location(self, locations: Locations):
        self.add_node_by_match_idx(nodes=locations, new_node_class=LocationEntry)
        return self
